/* MIDICCI - created on 7/9/2002 by Tom Grandgent (tom@grandgent.com)
 * Copyright (c) 2002 by Tom Grandgent
 *
 * MIDI Continuous Controller Input (Interface?)
 *
 * Last updated: 7/10/2002
 */

#define _WIN32_IE 0x0300	// for backwards compatibility with older Windows versions

#include <windows.h>		// for everything
#include <windowsx.h>		// for message crackers
#include <crtdbg.h>			// for _ASSERTE macro (readable debug-build assertion checking)
#include "tgRegistry.h"		// for registry access
#include "tgDBuffer.h"		// for double buffered drawing
#include "tgMIDI.h"			// for MIDI I/O
#include "MIDICCI.h"		// stuff too big to fit comfortably here
#include "resource.h"		// for resources (resource.h is auto-generated by VC++)

// Defines
#define WM_SHOW_WELCOME		WM_APP + 1	// welcome message displayed after window has been created
#define BASE_DEVICE_MENU	200			// base ID for device menu items

// Constants
char *regKeyName = "Software\\Tom Grandgent\\MIDICCI";	// Registry key used for program settings
const char *mainClassName = "MIDICCI";		// Class name registered for main window
const char *mainWindowTitle = "MIDICCI";	// Window title for main window

// Global variables
HINSTANCE hInstance;					// Instance handle of this program's image
HWND hwndApp;							// Window handle of this program's main window
HWND hwndCC1 = NULL, hwndCC2 = NULL;	// Window handles of CC combo boxes
int CC1, CC2;							// CC #'s
int CC1Value = 63, CC2Value = 63;		// CC values
tgMIDI *midi;							// MIDI I/O object
HFONT ctrlFont;							// Font used for combo boxes
int mouseCaptured = FALSE;				// Is the mouse currently captured?
HWND oldCapture;						// Old window that had captured the mouse (likely NULL)

// Function prototypes
int WINAPI WinMain(HINSTANCE hCurInstance, HINSTANCE hPrevInstance, LPSTR lpszCmdLine, int nCmdShow);
void saveWindowPlacement(HWND hwnd, char *windowName);
int loadWindowPlacement(HWND hwnd, char *windowName, int *x, int *y, int *w, int *h, int resize);

LRESULT CALLBACK WndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);
int RegisterWindowClasses(void);

// Program entrypoint
int WINAPI WinMain(HINSTANCE hCurInstance, HINSTANCE hPrevInstance, LPSTR lpszCmdLine, int nCmdShow)
{
	// Save the hInstance
	hInstance = hCurInstance;

	// Register window classes
	if (RegisterWindowClasses())
	{
		MessageBox(NULL, "Error registering window classes!", "Initialization error", MB_ICONERROR);
		return 1;
	}

	// Try to load window dimensions and location from the registry
	int x = (GetSystemMetrics(SM_CXSCREEN) - 325) / 2;
	int y = (GetSystemMetrics(SM_CYSCREEN) - 325) / 2;
	int w = 325, h = 325;
	loadWindowPlacement(NULL, NULL, &x, &y, &w, &h, TRUE);

	// Create the main window
	hwndApp = CreateWindow(mainClassName, mainWindowTitle, 
			   WS_OVERLAPPEDWINDOW | WS_VISIBLE, x, y, 
			   w, h, HWND_DESKTOP, NULL, hInstance, NULL);
	if (!hwndApp)
	{
		int err = GetLastError();
		char buf[256];
		sprintf(buf, "Error %d trying to create main application window!", err);
		MessageBox(NULL, buf, mainWindowTitle, MB_ICONERROR);
		return 1;
	}

	// Main program loop
	MSG msg;
	while (GetMessage(&msg, NULL, 0, 0))
	{
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}

	return msg.wParam;
}

// Save the location of a window in the registry (requires tgRegistry.cpp)
void saveWindowPlacement(HWND hwnd, char *windowName)
{
	RECT r;
	WINDOWPLACEMENT wp;
	char fullKeyName[256];

	strcpy(fullKeyName, regKeyName);
	if (windowName)
	{
		_ASSERTE(strlen(windowName) < 128);
		strcat(fullKeyName, "\\");
		strcat(fullKeyName, windowName);
	}

	wp.length = sizeof(wp);
	GetWindowPlacement(hwnd, &wp);
	memcpy(&r, &wp.rcNormalPosition, sizeof(RECT));
	if (r.right > r.left && r.bottom > r.top)
	{
		tgRegistry reg(fullKeyName);
		reg.setValueInt("windowX", r.left);
		reg.setValueInt("windowY", r.top);
		reg.setValueInt("windowWidth", r.right - r.left);
		reg.setValueInt("windowHeight", r.bottom - r.top);
		reg.setFlag("windowSizeSaved", TRUE);
	}
}

// Try to load window dimensions and location from the registry (requires tgRegistry.cpp)
// Which parameters should be passed is as follows:
// If hwnd is provided then no need for providing x, y, w, h
// If hwnd is not provided then x, y, w, h must be provided
// windowName may or may not be provided in any case
// resize determines whether the window should be just repositioned or repositioned and resized
int loadWindowPlacement(HWND hwnd, char *windowName, int *x, int *y, int *w, int *h, int resize)
{
	char fullKeyName[256];

	strcpy(fullKeyName, regKeyName);
	if (windowName)
	{
		_ASSERTE(strlen(windowName) < 128);
		strcat(fullKeyName, "\\");
		strcat(fullKeyName, windowName);
	}

	tgRegistry reg(fullKeyName);
	if (reg.getFlag("windowSizeSaved"))
	{
		// Behavior differs depending on whether or not a window handle was provided
		if (hwnd)
		{
			// Get the current window placement
			WINDOWPLACEMENT wp;
			wp.length = sizeof(wp);
			GetWindowPlacement(hwnd, &wp);
			// Save the old width and height
			int ow = wp.rcNormalPosition.right - wp.rcNormalPosition.left;
			int oh = wp.rcNormalPosition.bottom - wp.rcNormalPosition.top;
			// Update the normal dimensions from values saved in the registry
			reg.getValueInt("windowX", (int *) &wp.rcNormalPosition.left);
			reg.getValueInt("windowY", (int *) &wp.rcNormalPosition.top);
			if (resize)
			{
				reg.getValueInt("windowWidth", (int *) &wp.rcNormalPosition.right);
				reg.getValueInt("windowHeight", (int *) &wp.rcNormalPosition.bottom);
				wp.rcNormalPosition.right += wp.rcNormalPosition.left;
				wp.rcNormalPosition.bottom += wp.rcNormalPosition.top;
			}
			else
			{
				wp.rcNormalPosition.right = wp.rcNormalPosition.left + ow;
				wp.rcNormalPosition.bottom = wp.rcNormalPosition.top + oh;
			}
			// Set the new window placement
			SetWindowPlacement(hwnd, &wp);
		}
		else
		{
			// Load the values from the registry into the given variables
			reg.getValueInt("windowX", x);
			reg.getValueInt("windowY", y);
			reg.getValueInt("windowWidth", w);
			reg.getValueInt("windowHeight", h);
		}
		return 0;
	}

	return 1;
}

// WM_CLOSE: The window has been asked to close
void Main_OnClose(HWND hwnd)
{
	// Get rid of the MIDI I/O object
	delete midi;

	// Delete the font used for the combo boxes
	DeleteObject(ctrlFont);

	// Save the location of the main window in the registry
	saveWindowPlacement(hwnd, NULL);

	// Exit the program
	PostQuitMessage(0);
}

// Handles combo box notification messages
// Returns TRUE if a message was processed, otherwise FALSE
int Main_OnCommand_ComboBox(HWND hwnd, int id, HWND hwndCtl, UINT codeNotify)
{
	// Handle CBN_SELCHANGE messages (selection has changed)
	if (codeNotify == CBN_SELCHANGE)
	{
		tgRegistry reg(regKeyName);
		int sel = ComboBox_GetCurSel(hwndCtl);
		if (hwndCtl == hwndCC1)		// CC1 combo box
		{
			CC1 = sel - 1;
			reg.setValueInt("CC1", CC1);
			return TRUE;
		}
		else
		if (hwndCtl == hwndCC2)		// CC2 combo box
		{
			CC2 = sel - 1;
			reg.setValueInt("CC2", CC2);
			return TRUE;
		}
	}

	return FALSE;
}

// Builds/rebuilds the device menu
void Main_BuildDeviceMenu(HWND hwnd)
{
	// Get the menu handle for the Devices menu
	HMENU menu = GetMenu(hwnd);
	HMENU submenu = GetSubMenu(menu, 1);
	// Delete old menu items
	while (GetMenuItemCount(submenu))
		DeleteMenu(submenu, 0, MF_BYPOSITION);
	// Add the [none] item
	int i = 0;
	int deviceID = midi->getOutputDeviceID();
	AppendMenu(submenu, MF_STRING | ((deviceID == -1) ? MF_CHECKED : 0), 
		BASE_DEVICE_MENU + i++, "[none]");
	// Add the device items
	for (midi_device_t *d = midi->midi_devices; d; d = d->next)
		if (!d->input_device)
			AppendMenu(submenu, MF_STRING | ((d->deviceID == deviceID) ? MF_CHECKED : 0), 
				BASE_DEVICE_MENU + i++, d->name);
	// Regenerate the menu
	DrawMenuBar(hwnd);
}

// WM_COMMAND: A message from a control, menu, accelerator, and others
void Main_OnCommand(HWND hwnd, int id, HWND hwndCtl, UINT codeNotify)
{
	// Combo box messages
	if (Main_OnCommand_ComboBox(hwnd, id, hwndCtl, codeNotify))
		return;
	// Device menu messages
	if ((id >= BASE_DEVICE_MENU) && id < (BASE_DEVICE_MENU + 50))
	{
		if (id == BASE_DEVICE_MENU)
			midi->closeMidiOut();	// no device was selected
		else
			midi->openMidiOut(id - BASE_DEVICE_MENU - 1);	// a device was selected
		// Save the selection in the registry
		tgRegistry reg(regKeyName);
		reg.setValueInt("outputDeviceID", (int) (id - BASE_DEVICE_MENU - 1));
		Main_BuildDeviceMenu(hwnd);
		return;
	}

	// Menu messages
	switch (id)
	{
		case IDM_HELP_HELP:
			MessageBox(hwnd, 
				"This program lets you send MIDI control change events for one or two "
				"continuous controllers by dragging the mouse around on a two-dimensional grid.  "
				"First choose a MIDI-out device from the Device menu.  Then choose the controller(s) "
				"you want to adjust from the two combo boxes.  Now you're ready to drag the mouse "
				"around on the grid.  The first controller is adjusted along the horizontal axis "
				"and the second along the vertical axis.  Simple, eh?\n\n"
				"Sometimes it's nice to be able to control two controllers at once.  For example: "
				"CC74 (Sound Brightness/cutoff frequency) and CC71 (Sound Timbre/resonance).  Or, "
				"CC76 (Sound Vibrato Rate) and CC77 (Sound Vibrato Depth).  There are many "
				"possibilities, especially if you're able to customize the functions of two "
				"controllers for your own purposes, like adjusting insertion effect parameters.",
				mainWindowTitle, MB_ICONINFORMATION);
			break;

		case IDM_HELP_ABOUT:
			MessageBox(hwnd,
				"MIDICCI - MIDI Continuous Controller Input\n\n"
				"Written by Tom Grandgent on July 9th, 2002\n"
				"Send comments, questions, suggestions to tom@grandgent.com\n\n"
				"This program was written in C++ using the Win32 API (no MFC)", 
				mainWindowTitle, MB_ICONINFORMATION);
			break;

		case IDM_FILE_EXIT:
			PostMessage(hwnd, WM_CLOSE, 0, 0);
			break;
	}
}

// WM_CREATE: The window is being created
BOOL Main_OnCreate(HWND hwnd, LPCREATESTRUCT lpCreateStruct)
{
	char buf[256];
	int i;
	tgRegistry reg(regKeyName);

	// Create a font to use for controls
	HDC hdc = GetDC(hwnd);
	int nHeight = -MulDiv(8, GetDeviceCaps(hdc, LOGPIXELSY), 72);
	ReleaseDC(hwnd, hdc);
	ctrlFont = CreateFont(nHeight, 0, 0, 0, FW_NORMAL, 0, 0, 0, DEFAULT_CHARSET, 
		OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH, "MS Sans Serif");

	// Open MIDI output and/or input devices
	midi = new tgMIDI;
	reg.getValueInt("outputDeviceID", &i, -1);
	if (i != -1)
	{
		midi->openMidiOut(i);
	}

	// Create the CC1 combo box
	hwndCC1 = CreateWindow("COMBOBOX", "", CBS_DROPDOWNLIST | WS_VISIBLE | WS_CHILD | WS_VSCROLL, 
		0, 0, 140, 200, hwnd, (HMENU) 0, hInstance, NULL);
	SendMessage(hwndCC1, WM_SETFONT, (WPARAM) ctrlFont, MAKELPARAM(FALSE, 0));

	// Create the CC2 combo box
	hwndCC2 = CreateWindow("COMBOBOX", "", CBS_DROPDOWNLIST | WS_VISIBLE | WS_CHILD | WS_VSCROLL, 
		150, 0, 140, 200, hwnd, (HMENU) 0, hInstance, NULL);
	SendMessage(hwndCC2, WM_SETFONT, (WPARAM) ctrlFont, MAKELPARAM(FALSE, 0));

	// Fill the CC combo boxes
	ComboBox_AddString(hwndCC1, "[none]");
	ComboBox_AddString(hwndCC2, "[none]");
	for (i = 0; i < sizeof(controllerNames) / sizeof(char *); i++)
	{
		sprintf(buf, "%d:%s", i, controllerNames[i]);
		ComboBox_AddString(hwndCC1, buf);
		ComboBox_AddString(hwndCC2, buf);
	}
	reg.getValueInt("CC1", &CC1, 74);
	reg.getValueInt("CC2", &CC2, 71);
	ComboBox_SetCurSel(hwndCC1, CC1 + 1);
	ComboBox_SetCurSel(hwndCC2, CC2 + 1);

	// Fill in the device menu
	Main_BuildDeviceMenu(hwnd);

	// Display a welcome message on the first run
	if (!reg.getFlag("runBefore"))
	{
		reg.setFlag("runBefore", TRUE);
		PostMessage(hwnd, WM_SHOW_WELCOME, 0, 0);
	}

	return TRUE;
}

// WM_PAINT: Handles paint requests - paints the CC grid and the blank rect above it
void Main_OnPaint(HWND hwnd)
{
	PAINTSTRUCT ps;
	HDC hdc;
	int x, y;
	float f, inc;

	// Painting / double buffering / rect setup
	hdc = BeginPaint(hwnd, &ps);
	if (hdc)
	{
		// Create GDI objects
		HPEN centerPen = CreatePen(PS_SOLID, 1, RGB(0, 0, 255));
		HBRUSH centerBrush = CreateSolidBrush(RGB(0, 0, 127));

		// Create the double buffer object and get some rects
		static tgDBuffer db(hwnd);
		RECT r, cbRect, r2;
		GetClientRect(hwnd, &r);
		GetClientRect(hwndCC1, &cbRect);

		// Draw a filled gray rectangle to serve as a background for the combo box area
		r2.left = r.left;
		r2.right = r.right;
		r2.top = r.top;
		r2.bottom = cbRect.bottom;
		// Exclude any combo boxes n' stuff from the clipping rectangle to avoid flicker
		ExcludeClipRect(hdc, cbRect.left, cbRect.top, cbRect.right, cbRect.bottom);
		GetClientRect(hwndCC2, &cbRect);
		cbRect.left += 150;
		cbRect.right += 150;
		ExcludeClipRect(hdc, cbRect.left, cbRect.top, cbRect.right, cbRect.bottom);
		FillRect(hdc, &r2, (HBRUSH) GetStockObject(LTGRAY_BRUSH));

		// Get the double buffer and mess around with rects
		r.top += cbRect.bottom - cbRect.top;
		hdc = db.getBuffer(&r);
		r.top = 0;
		r.bottom -= cbRect.bottom - cbRect.top;

		// BEGIN DRAWING
		SelectObject(hdc, GetStockObject(BLACK_PEN));
		Rectangle(hdc, r.left, r.top, r.right, r.bottom);
		// Horizontal grid lines
		inc = (float) (r.bottom - r.top) / 8.0f;
		for (f = (float) r.top; f < (float) r.bottom; f += inc)
		{
			y = (int) f;
			MoveToEx(hdc, 0, y, NULL);
			LineTo(hdc, r.right, y);
		}
		// Vertical grid lines
		inc = (float) (r.right - r.left) / 8.0f;
		for (f = (float) r.left; f < (float) r.right; f += inc)
		{
			x = (int) f;
			MoveToEx(hdc, x, 0, NULL);
			LineTo(hdc, x, r.bottom);
		}

		// Center lines
		SelectObject(hdc, centerPen);
		// Horizontal center line
		MoveToEx(hdc, r.left, (r.bottom - r.top) / 2, NULL);
		LineTo(hdc, r.right, (r.bottom - r.top) / 2);
		// Vertical center line
		MoveToEx(hdc, (r.right - r.left) / 2, r.top, NULL);
		LineTo(hdc, (r.right - r.left) / 2, r.bottom);
		
		// CC data circle
		x = (int) (((float) CC1Value / 127.0f) * (float) (r.right - r.left));
		y = (int) (((float) (127 - CC2Value) / 127.0f) * (float) (r.bottom - r.top));
		SelectObject(hdc, centerBrush);
		Ellipse(hdc, x - 4, y - 4, x + 4, y + 4);
		// END OF DRAWING

		// Delete GDI objects
		SelectObject(hdc, GetStockObject(WHITE_BRUSH));
		DeleteObject(centerBrush);
		SelectObject(hdc, GetStockObject(BLACK_PEN));
		DeleteObject(centerPen);

		// Blit the double buffer to the screen
		db.draw();
	}
	EndPaint(hwnd, &ps);
}

// WM_MOUSEMOVE and WM_LBUTTONDOWN: Handles mouse clicks and movement on the CC grid
void Main_OnMouseMove(HWND hwnd, int x, int y, UINT keyFlags)
{
	if (!(keyFlags & MK_LBUTTON))
		return;

	// Capture the mouse, automatically handling border conditions nicely
	if (!mouseCaptured)
	{
		mouseCaptured = TRUE;
		oldCapture = SetCapture(hwnd);
	}

	// Figure out the CC grid rect
	RECT r, r2;
	GetClientRect(hwnd, &r);
	GetClientRect(hwndCC1, &r2);
	int cbHeight = r2.bottom - r2.top;
	r.top += cbHeight;
	
	// Calculate the CC values from the mouse position
	if (CC1 != -1)
	{
		CC1Value = (int) ((float) x / (float) (r.right - r.left) * 127.0f);
		if (CC1Value > 127)
			CC1Value = 127;
		else if (CC1Value < 0)
			CC1Value = 0;
	}
	else
		CC1Value = 63;
	if (CC2 != -1)
	{
		CC2Value = 127 - (int) ((float) (y - cbHeight) / (float) (r.bottom - r.top) * 127.0f);
		if (CC2Value > 127)
			CC2Value = 127;
		else if (CC2Value < 0)
			CC2Value = 0;
	}
	else
		CC2Value = 63;
	// Send control change events to the MIDI out device with the new CC values
	if (CC1 != -1)
		midi->controlChange(CC1, CC1Value, 0);
	if (CC2 != -1)
		midi->controlChange(CC2, CC2Value, 0);
	// Make sure the CC grid gets redrawn, so the new CC position is shown
	InvalidateRect(hwnd, NULL, FALSE);

	// Update the window title with the CC values
	char buf[256];
	strcpy(buf, mainWindowTitle);
	if (CC1 != -1)
		sprintf(buf + strlen(buf), " - CC%d=%d (%d)", CC1, CC1Value, CC1Value - 64);
	if (CC2 != -1)
		sprintf(buf + strlen(buf), " - CC%d=%d (%d)", CC2, CC2Value, CC2Value - 64);
	SetWindowText(hwnd, buf);
}

// WM_LBUTTONDOWN: Use the WM_MOUSEMOVE handler - we want same functionality for both msgs
void Main_OnLButtonDown(HWND hwnd, BOOL fDoubleClick, int x, int y, UINT keyFlags)
{
	Main_OnMouseMove(hwnd, x, y, keyFlags);
}

// Main window procedure, processes window messages
LRESULT CALLBACK WndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch (msg)
	{
		HANDLE_MSG(hwnd, WM_CREATE, Main_OnCreate);
		HANDLE_MSG(hwnd, WM_COMMAND, Main_OnCommand);
		HANDLE_MSG(hwnd, WM_CLOSE, Main_OnClose);
		HANDLE_MSG(hwnd, WM_PAINT, Main_OnPaint);
		HANDLE_MSG(hwnd, WM_MOUSEMOVE, Main_OnMouseMove);
		HANDLE_MSG(hwnd, WM_LBUTTONDOWN, Main_OnLButtonDown);

		case WM_LBUTTONUP:
			// Release mouse capture on mouse up
			if (mouseCaptured)
			{
				mouseCaptured = FALSE;
				SetCapture(oldCapture);
			}
			break;

		case WM_SHOW_WELCOME:
			MessageBox(hwnd, 
				"Welcome to MIDICCI.\n\nTo get started, select a MIDI-out device from the "
				"Device menu, \nchoose controllers, and drag the mouse around on the grid.\n\n"
				"More information is available in the Help menu.", 
				mainWindowTitle, MB_ICONINFORMATION);
			break;
	}
	return DefWindowProc(hwnd, msg, wParam, lParam);
}

// Registers window classes for windows created by this program
int RegisterWindowClasses(void)
{
	WNDCLASSEX wcl;

	// Initialize the main window class
	wcl.cbSize = sizeof(WNDCLASSEX);
	wcl.style = CS_HREDRAW | CS_VREDRAW;
	wcl.lpfnWndProc = WndProc;
	wcl.cbClsExtra = 0;
	wcl.cbWndExtra = 0;
	wcl.hInstance = hInstance;
	wcl.hIcon = LoadIcon(hInstance, MAKEINTRESOURCE(IDI_MAIN));
	wcl.hCursor = LoadCursor(NULL, IDC_ARROW);
	wcl.hbrBackground = 0; //(HBRUSH) GetStockObject(LTGRAY_BRUSH);
	wcl.lpszMenuName = MAKEINTRESOURCE(IDM_MAIN);
	wcl.lpszClassName = mainClassName;
	wcl.hIconSm = (HICON) LoadImage(hInstance, MAKEINTRESOURCE(IDI_MAIN), IMAGE_ICON, 
					GetSystemMetrics(SM_CXSMICON),
					GetSystemMetrics(SM_CYSMICON), 0);
	if (!RegisterClassEx(&wcl))
		return 1;

	return 0;
}
